#include "CcbrEngine.h"

namespace CCBR_Module{

	CcbrEngine::CcbrEngine(CaseBase* casebase, string solutionFeature, double applicationSimThreshold, double dialogueSimThreshold):casebase_(casebase), solutionFeature_(solutionFeature), similarityThreshold_(applicationSimThreshold){
		dialogueSystem_ = new DialogueSystem(casebase_, dialogueSimThreshold);
		queryCounter_ = 0;
		dl_ = true; // default enabled
	}

	CcbrEngine::~CcbrEngine(){}

	bool CcbrEngine::query(vector<pair<string, string> > problemDescription, Agent* agent){
		cout << "CCBR-> New query started." << endl;
		features_ = casebase_->getFeatureTypes();
		// Solution should not be asked for as a question
		features_.erase(casebase_->getEntity(solutionFeature_));

		vector<pair<Case*, double> > solutions;
		vector<pair<string, double> > questions;

		QueryBiasedSymbolicComparator sc;
		SimilarityMeasureCaseRetriever crsc(casebase_, &sc, similarityThreshold_);

		Case* query = new Case(generateId(++queryCounter_), casebase_);

		// Insert problem description / remove questions candidates from features
		for (unsigned int i = 0; i < problemDescription.size(); i++) {
			query->addFeature(problemDescription[i].first, problemDescription[i].second);
			features_.erase(casebase_->getEntity(problemDescription[i].first));
		}

		// Create new dialogue session
		if(dl_) dialogueSystem_->newSession(query);

		// Start ccbr cycle
		Case* solution = NULL;
		while(solution == NULL) {
			cout << "CCBR-> Retrieve similar cases." << endl;
			solutions = crsc.retrieveSimilarCases(query);

			cout << "CCBR-> Generate ranked questions." << endl;
			questions = generateQuestions();
			//questions = generateQuestions(query, solutions);

			// Interact with user/agent
			if(questions.empty()) {
				cout << "CCBR-> No more questions available." << endl;
				if(dl_)	dialogueSystem_->discardSession();
				delete query;
				return false;
			} else {
				// Improve questions and solutions with dialogue session
				if(dl_) dialogueSystem_->reuseDialogue(&solutions, &questions);

				// If there are more questions to ask, ask the agent
				pair<string, string> newFeature = agent->askQuestions(&questions);
				if (newFeature.first == "") { // The user could not answer any more questions
					if(dl_)	dialogueSystem_->discardSession();
					delete query;
					return false;
				}
				// Register the answer
				features_.erase(casebase_->getEntity(newFeature.first));
				query->addFeature(newFeature.first, newFeature.second);

				// Notify the dialogue system
				if(dl_)	dialogueSystem_->addFeature(newFeature.first);
			}

			if(solutions.size() > 0)
				solution = agent->askSolutions(&solutions);
		}
		// Retain the query with solution
		//Feature* fs = solution->getFeature(solutionFeature_);
		//query->addFeature(solutionFeature_, fs->getValue()->getTextValue());
		delete query;  // want the case base to be stabile during testing, no new application cases.


		// Commit the dialogue session
		if(dl_)	dialogueSystem_->commitSession(solution);

		return true;
	}

	struct qwpair
	{
		bool operator()(const pair<string, double>& a, const pair<string, double>& b) const {
			return a.second > b.second;
		}
	};

	vector<pair<string, double> > CcbrEngine::generateQuestions() {
		vector<pair<string, double> > questions;

		// Rank remaining features
		for (set<Entity*>::const_iterator it = features_.begin(); it != features_.end(); ++it) {
			Entity* e = *it;
			questions.push_back(make_pair(e->getName(), casebase_->getWeight(e->getName())));
		}

		sort(questions.begin(), questions.end(), qwpair());
		return questions;
	}

	// Using information gain.
	vector<pair<string, double> > CcbrEngine::generateQuestions(Case* query, vector<pair<Case*, double> > caseSimilarities) {
		cout << "CCBR-> Generating questions." << endl;
		vector<pair<string, double> > questions;
		vector<pair<Entity*, double> > ranking;

		// Ranking done by Ida Kokkersvold
		InformationGain* infogain = new InformationGain();
		ranking = infogain->rankFeatures(query, solutionFeature_, caseSimilarities, casebase_);

		for (vector<pair<Entity*, double> >::const_iterator it = ranking.begin(); it != ranking.end(); ++it) {
			pair<Entity*, double> p = *it;
			questions.push_back(make_pair(p.first->getName(), p.second));
		}

		sort(questions.begin(), questions.end(), qwpair());
		return questions;

	}

	void CcbrEngine::enableDialogueLearning() {
		dl_ = true;
		cout << "CCBR-> Dialogue learning enabled." << endl;
	}

	void CcbrEngine::disableDialogueLearning() {
		dl_ = false;
		cout << "CCBR-> Dialogue learning disabled." << endl;
	}

	int CcbrEngine::getNumberOfStoredDialogues() {
		return dialogueSystem_->getNumberOfStoredDialogues();
	}

}
