#include "DialogueSystem.h"
#include <cassert>

namespace CCBR_Module
{

	DialogueSystem::DialogueSystem(CaseBase* applicationCaseBase, double similarityThreshold):applicationCaseBase_(applicationCaseBase), similarityThreshold_(similarityThreshold){
		dialogueCaseBase_ = new CaseBase("DialogueCaseBase");
		sessionNumber_ = 0;
	}

	DialogueSystem::~DialogueSystem(){}

	void DialogueSystem::newSession(Case* initialAppCase) {
		cout << "DS-> New session started." << endl;

		currentAppCase_ = initialAppCase;

		// Create new case to store dialogue session
		currentPosition_ = 0;
		currentSession_ = new Case(generateDialogueId(++sessionNumber_),dialogueCaseBase_);

/*
		// Add the problemdescription as the first steps of the dialogue
		vector<Feature*> pd = currentAppCase_->getFeatures();
		for (unsigned int i = 0; i < pd.size(); i++) {
			currentSession_->addFeature(pd[i]->getName(), convertToString(++currentPosition_));
		}
*/
		cout << "DS-> Current dialogue base size: " << dialogueCaseBase_->getCases().size() << endl;
	}

	void DialogueSystem::reuseDialogue(vector<pair<Case*, double> >* solutions, vector<pair<string, double> >* questions) {
		assert(questions);
		assert(solutions);
		DialogueCaseComparator dcc;
		DialogueCaseRetriever dialogueRetriever(dialogueCaseBase_, applicationCaseBase_, &dcc, similarityThreshold_);

		vector<pair<Case*, double> > dialogues = dialogueRetriever.retrieveSimilarCases(currentSession_, currentAppCase_);

		if(dialogues.size()>0) {
			cout << "DS-> Best: " << dialogues[0].first->getName() << " (" << dialogues[0].second << ")" << endl;

			// Get best dialogue case
			Case* bestDialogue = dialogues[0].first;

			// Replace last solution by the solution in the best dialogue solution
			// (if not included already)
			Feature* solutionFeature = bestDialogue->getFeature("ds_solution");
			assert(solutionFeature);
			Case* solutionAppCase = dynamic_cast<Case*>(applicationCaseBase_->getEntity(solutionFeature->getValue()->getTextValue()));

			bool solutionIncluded = false;
			for (unsigned int i = 0; i < solutions->size() && !solutionIncluded; i++) {
				if(solutions->at(i).first == solutionAppCase) {
					solutionIncluded = true;
				}
			}

			if(!solutionIncluded) {
				solutions->push_back(make_pair(solutionAppCase, dialogues[0].second));
			}

			// Update weights for features included in the best dialogue case.
			double featureCount = applicationCaseBase_->getFeatureTypes().size()-1;  // not counting the solution feature
			double dialogueLength = bestDialogue->getFeatures().size();

			for (unsigned int i = 0; i < questions->size(); i++) {
				if(bestDialogue->hasFeature(questions->at(i).first)) {
					double featurePosition = convertToInt(bestDialogue->getFeature(questions->at(i).first)->getValue()->getTextValue());
					double weight = questions->at(i).second + (0.5 + 0.5 * (1.0 - featurePosition / dialogueLength)) * (1.0 / featureCount);
					cout << "DS-> Weight for " << questions->at(i).first << " changed from "<< questions->at(i).second << " to " << weight << endl;
					questions->at(i).second = weight;
				}
			}
		} else {
			cout << "DS-> No dialogue cases better than threshold value." << endl;
		}
	}

	void DialogueSystem::addFeature(string featureType) {
		// Add the featuretype with the corresponding position in the dialogue
		// no need to store values, as they can be obtained form the application case base
		currentSession_->addFeature(featureType, convertToString(++currentPosition_));
	}

	void DialogueSystem::discardSession() {
		// If the ccbr cycle is not successful the session should not be saved.
		//cout << "Delete dialogue..." << endl;
		delete currentSession_;
		//cout << "Deleted dialogue." << endl;
	}

	bool DialogueSystem::commitSession(Case* applicationCase) {
		// If the ccbr cycle is successful the application case is added as the solution
		// for the session, then the session is considered for retain.
		currentSession_->addFeature("ds_solution", applicationCase->getName());

		QueryBiasedSymbolicComparator sc;
		SimilarityMeasureCaseRetriever generalRetriever(dialogueCaseBase_, &sc, 1.0);

		vector<pair<Case*, double> > moreSpecializedCases = generalRetriever.retrieveSimilarCases(currentSession_);

		for(unsigned int i=0; i<moreSpecializedCases.size(); i++) {
			cout << "DS-> Removed " << moreSpecializedCases[i].first->getName() << " as a result of a more general case stored." << endl;
			delete moreSpecializedCases[i].first;
		}

		cout << "DS-> Session retained." << endl;
		// return true is session is retained, or else false.
		return true;
	}

	int DialogueSystem::getNumberOfStoredDialogues() {
		return dialogueCaseBase_->getActiveCases().size();
	}
}
